#Copyright 2006 DR0ID <dr0id@bluewin.ch> http://mypage.bluewin.ch/DR0ID
#
#
#
"""
#TODO: documentation!
"""

__author__   = "$Author: DR0ID $"
__version__  = "$Revision: 37 $"
__date__     = "$Date: 2006-10-18 14:35:04 +0200 (Mi, 18 Okt 2006) $"
__license__  = ''
__copyright__= "DR0ID (c) 2006"

import pygame
import math
import os

## TODO:- abstract physics model (ship inherits)
##      - layers?
##      - dirty?
##      - proper coord transforms
##      - resoucemanager
##      - map
##      - animation for thrusters??


def rotate_point(point, rad_angle):
    px = point[0]
    py = point[1]
    cosv = math.cos(rad_angle)
    sinv = math.sin(rad_angle)
    return ( px*cosv-py*sinv, px*sinv+py*cosv)


class G(pygame.sprite.OrderedUpdates):
    
    def draw(self, surface):
##        spritedict = self.spritedict
        surface_blit = surface.blit
##       dirty = self.lostsprites
##       self.lostsprites = []
##       dirty_append = dirty.append
        wrap = 0
        for s in self.sprites():
            #TODO: use if s.visible: here and make object use visible (or dirty?)
##            r = spritedict[s]
            r = s.rect
            newrect = surface_blit(s.image, r)
            if r.left<0:
                r.move_ip(800, 0)
                wrap = 1
            elif r.right > 800:
                r.move_ip(-800, 0)
                wrap = 1
            if r.top < 0:
                r.move_ip(0, 600)
                wrap = 1
            elif r.bottom > 600:
                r.move_ip(0, -600)
                wrap = 1
            if wrap:
                surface_blit(s.image, r)
                wrap = 0
        
##            spritedict[s] = newrect


class Thruster(pygame.sprite.Sprite):
    
    def __init__(self, ship, attached_where, direction, thrust, gpos, gscale, images):
        '''
        scale: scale of thruster (does not affect thrust)
        thrust: force of thruster
        '''
        pygame.sprite.Sprite.__init__(self)
        # grahics
        self._images = images
        self._gscale = gscale
        self._gpos = (gscale*gpos[0], gscale*gpos[1])
        self.image, self.rect = images[0]
        self.rect = pygame.Rect(-10,-10,0,0)
        # state
        self.active = 0
        # ship
        self._ship = ship
        # vectors
        self._attached_where = attached_where # [m] local coord of the ship
        self._direction      = direction      # direction in which the thrust affects
        self._angle_offset   = 0              # after _calc angle offset to ship
        self._torque         = 0              # after _calc [Nm] it should be valid raound cg in (0,0,1)
        self._force          = (0,0)          # after _calc [N] force going through cg
        # thrust
        self._thrust = thrust #0 # [N] (magnitude of F)
        self._calc()
        
    def _calc(self):
        # (pre)calc force and torque 
        f = self._direction
        af = math.sqrt(f[0]*f[0]+f[1]*f[1]) # length of f
        f = (f[0]/af, f[1]/af) # normalized f unitvector
        f = (f[0] * self._thrust, f[1] * self._thrust)  # actual force [N]
        r = self._attached_where
        self._torque = -(r[0]*f[1]-r[1]*f[0]) # torque looking in z [Nm]
        # projection u onto v: p = dot(u,v)/dot(v,v)*v
        # project f onto r 
        s = (f[0]*r[0]+f[1]*r[1])/(r[0]*r[0]+r[1]*r[1])
        self._force = (s*r[0], s*r[1])
        self._angle_offset   = -int(round(math.degrees(math.atan2( self._direction[1], self._direction[0] ) ) ) )# because ship look at (1,0)
        print "thruster: force", self._force, "torque", self._torque, "angle", self._angle_offset
        
    def update(self, *args):
        if self.active:
            f = rotate_point(self._force, -self._ship.angle)
            self._ship.apply_force_cg(f)
            self._ship.apply_torque(self._torque)
            
            angle  = (args[1]+self._angle_offset)%360
            self.image, rect = self._images[angle]
            self.rect = rect.move(self._ship.pos)
            self.rect.move_ip( rotate_point(self._gpos, -self._ship.angle) )
        else:
            self.rect = pygame.Rect(-1000,-1000,0,0)




class Ship(pygame.sprite.Sprite):
    
    def __init__(self, scale, images):
        pygame.sprite.Sprite.__init__(self)
        # mass
        self.m      = 1000.0 #[kg]
        # velocity
        self.vx     = 0    #[m/s]
        self.vy     = 0
        # angle
        self.angle  = 0 #[radians]
        # angle velocity
        self.angleV = 0
        # torque
        self.torque = 0
        # force accumulation
        self.Fx     = 0    #[N]
        self.Fy     = 0
        # direction vector
        self.dx     = 1    
        self.dy     = 0
        # position
        self.pos    = [400,300]
        
        self.minVel = 0.00001
        # graphic stuff
        self._images       = images # 360
        self.image, self.rect = self._images[0]
        # state
        self.thruster_active = 0
        self.turning = 0
        # thrusters
        self._th_l_f  = None
        self._th_r_f  = None
        self._th_l_b  = None
        self._th_r_b  = None
        self._th_main = None
        
        self._thrusters = []
        
    def apply_force_cg(self, force):
        # add forces to one, all forces apply to point of gravtation
        # (if they would not apply to same point, then it would have to be
        # calculatet in a different way
        self.Fx += force[0]
        self.Fy += force[1]
        
    def apply_torque(self, amomentum):
        self.torque += amomentum

    def update(self, *args):
        dt = args[0]/1000.
        
        self._th_main.active = self.thruster_active
        if self.turning<0:
            self._th_l_f.active  = 1
            self._th_r_b.active  = 1
        elif self.turning>0:
            self._th_r_f.active  = 1
            self._th_l_b.active  = 1
        else:
            self._th_l_f.active  = 0
            self._th_r_f.active  = 0
            self._th_l_b.active  = 0
            self._th_r_b.active  = 0
            
        # angular things
        self.angleV += (self.torque/1000.)*dt
##        if abs(self.angleV)<self.minVel:
##            self.angleV = 0.0
        self.torque = 0
        self.angle += (self.angleV * dt)
        self.angle %= (2*math.pi)
        
        self.dx =  math.cos(self.angle)
        self.dy =  -math.sin(self.angle)
        
        # position things
        # acceleration
        ax = self.Fx/self.m
        ay = self.Fy/self.m
        # reset force
        self.Fx = 0
        self.Fy = 0
        # update position
        self.vx += ax * dt
        self.vy += ay * dt
##        if abs(self.vx) < self.minVel:
##            self.vx = 0.0
##        if abs(self.vy) < self.minVel:
##            self.vy = 0.0
        self.pos[0] += (self.vx  * dt)
        self.pos[1] += (self.vy  * dt)
        

        # wrapping
        # TODO: not really needed if using map
        self.pos[0] %= 800
        self.pos[1] %= 600
        
        # draw preparation
        angle = int(math.degrees(self.angle))%360
        self.image, rect = self._images[angle]
        self.rect = rect.move(self.pos)
        
        for t in self._thrusters:
            t.update(dt, angle)

    def attach_thruster(self, attr_name, th):
        self._thrusters.append(th)
        setattr(self, attr_name, th)



def main():
    
    # pygame init
    pygame.init()
    scr = pygame.display.set_mode((800,600))
    
    # scale of the ship
    scale = 0.5
    
    
    # load images
    # load and prepare them here (rotate)
    path, fname   = os.path.split(__file__)
    path          = os.path.join(path, 'data')
    bgd           = pygame.image.load(os.path.join(path, 'space.png')).convert()
    ship_img      = pygame.image.load(os.path.join(path, 'ship.png'))
    thruster_img  = pygame.image.load(os.path.join(path, 'thruster.png'))
    thruster_img2 = pygame.image.load(os.path.join(path, 'thruster2.png'))
    
    ship_img      = ship_img.convert()
    ship_img.set_colorkey((255,0,255))
    thruster_img  = thruster_img.convert()
    thruster_img.set_colorkey((255,0,255))
    thruster_img2 = thruster_img2.convert()
    thruster_img2.set_colorkey((255,0,255))
    
    print "preparing images..."
    w,h           = ship_img.get_size()
    ship_img      = pygame.transform.scale(ship_img, (scale*w, scale*h))
    w,h           = thruster_img.get_size()
    thruster_img  = pygame.transform.scale(thruster_img, (scale*w, scale*h))
    w,h           = thruster_img2.get_size()
    thruster_img2 = pygame.transform.scale(thruster_img2, (scale*w, scale*h))
    ship_images      = []
    thruster_images  = []
    thruster2_images = []
    for a in range(360):
        # ship
        img = pygame.transform.rotate(ship_img, a)
        r = img.get_rect(center=(0,0))
##        print a, r, r.center
        ship_images.append( (img,r) )
        
        # main thruster
        img = pygame.transform.rotate(thruster_img, a)
        r = img.get_rect(center=(0,0))
        offset = r.move( rotate_point((-42*scale,0*scale), -math.radians(a)) )
        thruster_images.append( (img,offset) )
        
        # steer thruster
        img = pygame.transform.rotate(thruster_img2, a)
        r = img.get_rect(center=(0,0))
        offset = r.move(rotate_point((-8*scale,0*scale), -math.radians(a) ))
        thruster2_images.append( (img, offset) )
        
    print "done"
    
    del img
    del r
    del offset
    print len(ship_images), len(thruster_images), len(thruster2_images)
        
    ship = Ship(scale, ship_images)
    ship.attach_thruster('_th_l_f', Thruster(ship, ( 6,-2.8), (0, 1),  100, ( 60,-30),scale, thruster2_images))
    ship.attach_thruster('_th_r_f', Thruster(ship, ( 6, 2.8), (0,-1),  100, ( 60, 30),scale, thruster2_images))
    ship.attach_thruster('_th_l_b', Thruster(ship, (-6,-2.8), (0, 1),  100, (-50,-66), scale, thruster2_images))
    ship.attach_thruster('_th_r_b', Thruster(ship, (-6, 2.8), (0,-1),  100, (-50, 66), scale, thruster2_images))
    ship.attach_thruster('_th_main', Thruster(ship, (-4,   0), (1, 0), 100000, (-40, 0), scale, thruster_images))
    
    g = G() #pygame.sprite.OrderedUpdates()
    g.add(ship)
    g.add(ship._thrusters)
    
    # main loop
    clock = pygame.time.Clock()
    looping = True
    fpslist = []
    timedelatation = 1
    while looping:
        # eventhandling
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                looping = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_UP:
                    ship.thruster_active = 1
                elif event.key == pygame.K_RIGHT:
                    ship.turning = -1
                elif event.key == pygame.K_LEFT:
                    ship.turning = 1
            elif event.type == pygame.KEYUP:
                if event.key == pygame.K_ESCAPE:
                    looping = False
                elif event.key == pygame.K_UP:
                    ship.thruster_active = 0
                elif event.key == pygame.K_LEFT or event.key == pygame.K_RIGHT:
                    ship.turning = 0
                    
        
        # timing things
        dt = clock.tick()
        dt = min(30, dt)
        fpslist.append(clock.get_fps())
##        pygame.display.set_caption(str(clock.get_fps())+" v = ("+str(ship.vx)+","+str(ship.vy)+") av="+str(ship.angleV))
        # draw things
        ship.update(dt)
        g.draw(scr)
        # fullscreen update
        pygame.display.flip()
##        pygame.display.update()
        scr.blit(bgd, (0,0))
        
    # finished the main loop, clean up
    print "avg fps: ", sum(fpslist)/float(len(fpslist))
    print "max fps: ", max(fpslist)
    fpsmin = 1000
    for i in fpslist:
        if i:
            if i<fpsmin:
                fpsmin=i
            
    print "min fps: ", fpsmin
    pygame.quit()    


if __name__ == '__main__':
##    main()
    import profile
    profile.run('main()')